#pragma once

#include <cstdint>

namespace AMDiS
{
  /** \ingroup Common
   * \brief
   * The Flag class encapsulates flags which represents simple information.
   * Used e.g. while mesh traversal to specify, which elements should be
   * visited.
   */
  class Flag
  {
  public:
    /// Constructs a unset Flag
    constexpr Flag() = default;

    /// Constructs a Flag initialized by f
    constexpr Flag(const std::uint64_t f)
      : flags(f)
    {}

    /// Copy constructor
    constexpr Flag(Flag const&) = default;

    /// Move constructor
    constexpr Flag(Flag&&) = default;

    /// Compares two Flags
    constexpr bool operator==(Flag const& f) const
    {
      return (flags == f.flags);
    }

    /// Compares two Flags
    constexpr bool operator!=(Flag const& f) const
    {
      return !(f == *this);
    }

    /// Assignment operator
    constexpr Flag& operator=(Flag const& f)
    {
      if (this != &f)
        flags = f.flags;
      return *this;
    }

    /// Typecast
    constexpr operator bool() const
    {
      return isAnySet();
    }

    /// Set \ref flags
    constexpr void setFlags(const std::uint64_t f)
    {
      flags = f;
    }

    /// Set \ref flags
    constexpr void setFlags(Flag const& f)
    {
      flags = f.flags;
    }

    /// Sets \ref flags to \ref flags | f
    constexpr void setFlag(const std::uint64_t f)
    {
      flags |= f;
    }

    /// Sets \ref flags to \ref flags | f.flags
    constexpr void setFlag(Flag const& f)
    {
      flags |= f.flags;
    }

    /// Sets \ref flags to \ref flags & ~f
    constexpr void unsetFlag(const std::uint64_t f)
    {
      flags &= ~f;
    }

    /// Sets \ref flags to \ref flags & ~f.flags
    constexpr void unsetFlag(Flag const& f)
    {
      flags &= ~f.flags;
    }

    constexpr std::uint64_t getFlags() const
    {
      return flags;
    }

    /// Returns \ref flags | f.flags
    friend Flag operator+(Flag r, Flag const& f)
    {
      r.setFlag(f);
      return r;
    }

    /// Returns \ref flags & ~f.flags
    friend Flag operator-(Flag r, Flag const& f)
    {
      r.unsetFlag(f);
      return r;
    }

    /// Returns \ref flags | f.flags
    friend Flag operator|(Flag r, Flag const& f)
    {
      r.setFlag(f);
      return r;
    }

    /// Returns \ref flags & f.flags
    friend Flag operator&(Flag r, Flag const& f)
    {
      r.flags &= f.flags;
      return r;
    }

    /// Sets \ref flags to \ref flags &= f.flags
    constexpr Flag operator&=(Flag const& f)
    {
      flags &= f.flags;
      return *this;
    }

    /// Returns \ref flags ^ f.flags
    friend Flag operator^(Flag r, Flag const& f)
    {
      r.flags ^= f.flags;
      return r;
    }

    /// Sets \ref flags to \ref flags & f.flags
    constexpr Flag& operator|=(Flag const& f)
    {
      if (this != &f)
        flags |= f.flags;
      return *this;
    }

    /// Returns ~\ref flags
    constexpr Flag operator~() const
    {
      Flag r;
      r.flags = ~flags;
      return r;
    }

    /// Checks whether all set bits of f.flags are set in \ref flags too.
    constexpr bool isSet(Flag const& f) const
    {
      return ((flags&f.flags) == f.flags);
    }

    /// Returns !\ref isSet(f)
    constexpr bool isUnset(Flag const& f) const
    {
      return !isSet(f);
    }

    /// Returns true if \ref flags != 0
    constexpr bool isAnySet() const
    {
      return (flags != 0);
    }

  protected:
    /// Internal flag representation
    std::uint64_t  flags = 0;
  };

} // end namespace AMDiS
