#pragma once

/** \file LinearSolverInterface.h */

/**
 * \defgroup Solver Solver module
 * @{ <img src="solver.png"> @}
 *
 * \brief
 * Contains all classes needed for solving linear and non linear equation
 * systems.
 */

#include <amdis/common/Utility.hpp>

#include <amdis/linear_algebra/PreconditionerInterface.hpp>
#include <amdis/linear_algebra/RunnerInterface.hpp>
#include <amdis/linear_algebra/SolverInfo.hpp>

namespace AMDiS
{
  /// Abstract base class for linear solvers
  template <class Matrix, class VectorX, class VectorB = VectorX>
  class LinearSolverInterface
  {
  protected:
    using RunnerBase = RunnerInterface<Matrix, VectorX, VectorB>;

  public:
    /// Destructor.
    virtual ~LinearSolverInterface() = default;

    /// Public method to call in order to solve a linear system Ax = b.
    /**
     * The method redirects to the specific linear solver and prints statistics
     * and error estimations at the end.
     *
     * The parameters correspond to
     *  \p A     A [block-]matrix that represents the system-matrix.
     *  \p x     A [block-]vector for the unknown components.
     *  \p b     A [block-]vector for the right-hand side of the linear system.
     **/
    void solve(Matrix const& A, VectorX& x, VectorB const& b,
               SolverInfo& solverInfo)
    {
      solveImpl(A, x, b, solverInfo);
    }

    // return the runner/worker corresponding to this solver (optional)
    virtual std::shared_ptr<RunnerBase> runner() { return {}; };

  private:
    /// main methods that all solvers must implement
    virtual void solveImpl(Matrix const& A, VectorX& x, VectorB const& b,
                           SolverInfo& solverInfo) = 0;
  };


  template <class Matrix, class Vector>
  struct SolverCreator;

} // end namespace AMDiS
