#pragma once

#include <boost/numeric/mtl/vector/dense_vector.hpp>

#include <dune/common/ftraits.hh>

#include <amdis/Output.hpp>
#include <amdis/linear_algebra/DOFVectorBase.hpp>

namespace AMDiS
{
  /// The basic container that stores a base vector and a corresponding basis
  template <class ValueType>
  class MtlVector
  {
  public:
    /// The type of the elements of the DOFVector
    using value_type = ValueType;

    /// The type of the elements of the DOFVector
    using block_type = ValueType;

    /// The underlying field type
    using field_type = typename Dune::FieldTraits<ValueType>::field_type;

    /// The type of the base vector
    using BaseVector = mtl::vec::dense_vector<ValueType>;

    /// The index/size - type
    using size_type  = typename BaseVector::size_type;

  public:
    /// Constructor. Constructs new BaseVector.
    MtlVector() = default;

    /// Return the data-vector \ref vector_
    BaseVector const& vector() const
    {
      return vector_;
    }

    /// Return the data-vector \ref vector_
    BaseVector& vector()
    {
      return vector_;
    }

    /// Return the current size of the \ref vector_
    size_type size() const
    {
      return mtl::vec::size(vector_);
    }

    /// Resize the \ref vector_ to the size \p s
    void resize(size_type s)
    {
      vector_.change_dim(s);
    }


    /// Access the entry \p i of the \ref vector with read-access.
    value_type const& operator[](size_type i) const
    {
      test_exit_dbg(i < mtl::vec::size(vector_),
        "Index {} out of range [0,{})", i, mtl::vec::size(vector_));
      return vector_[i];
    }

    /// Access the entry \p i of the \ref vector with write-access.
    value_type& operator[](size_type i)
    {
      test_exit_dbg(i < mtl::vec::size(vector_),
        "Index {} out of range [0,{})", i, mtl::vec::size(vector_));
      return vector_[i];
    }

    void set(field_type value)
    {
      vector_ = value;
    }

  private:
    /// The data-vector (can hold a new BaseVector or a pointer to external data
    BaseVector vector_;
  };


  template <class BasisType, class ValueType = double>
  struct DOFVector : public DOFVectorBase<BasisType, MtlVector<ValueType>>
  {
    using Super = DOFVectorBase<BasisType, MtlVector<ValueType>>;
    using Super::operator=;
    DOFVector(BasisType const& basis)
      : Super(basis)
    {}
  };


  /// Constructor a dofvector from given basis and name
  template <class ValueType = double, class Basis>
  DOFVector<Basis, ValueType>
  makeDOFVector(Basis const& basis)
  {
    return {basis};
  }

} // end namespace AMDiS
